import os
import pickle
import sys
from importlib.abc import Loader
from importlib.util import module_from_spec, spec_from_file_location
from typing import Optional, cast


def process_worker() -> None:
    # Redirect standard streams to os.devnull so that user code won't interfere with the
    # parent-worker communication
    stdin = sys.stdin
    stdout = sys.stdout
    sys.stdin = open(os.devnull)
    sys.stdout = open(os.devnull, 'w')

    stdout.buffer.write(b'READY\n')
    while True:
        retval = exception = None
        try:
            command, *args = pickle.load(stdin.buffer)
        except EOFError:
            return
        except BaseException as exc:
            exception = exc
        else:
            if command == 'run':
                func, args = args
                try:
                    retval = func(*args)
                except BaseException as exc:
                    exception = exc
            elif command == 'init':
                main_module_path: Optional[str]
                sys.path, main_module_path = args
                del sys.modules['__main__']
                if main_module_path:
                    # Load the parent's main module but as __mp_main__ instead of __main__
                    # (like multiprocessing does) to avoid infinite recursion
                    try:
                        spec = spec_from_file_location('__mp_main__', main_module_path)
                        if spec and spec.loader:
                            main = module_from_spec(spec)
                            cast(Loader, spec.loader).exec_module(main)
                            sys.modules['__main__'] = main
                    except BaseException as exc:
                        exception = exc

        try:
            if exception is not None:
                status = b'EXCEPTION'
                pickled = pickle.dumps(exception, pickle.HIGHEST_PROTOCOL)
            else:
                status = b'RETURN'
                pickled = pickle.dumps(retval, pickle.HIGHEST_PROTOCOL)
        except BaseException as exc:
            exception = exc
            status = b'EXCEPTION'
            pickled = pickle.dumps(exc, pickle.HIGHEST_PROTOCOL)

        stdout.buffer.write(b'%s %d\n' % (status, len(pickled)))
        stdout.buffer.write(pickled)

        # Respect SIGTERM
        if isinstance(exception, SystemExit):
            raise exception


if __name__ == '__main__':
    process_worker()



import typing
from abc import ABCMeta, abstractmethod
from types import TracebackType
from typing import Callable, Coroutine, Optional, Type, TypeVar

from anyio._core._compat import DeprecatedAwaitable

if typing.TYPE_CHECKING:
    from anyio._core._tasks import CancelScope

T_Retval = TypeVar('T_Retval')


class TaskStatus(metaclass=ABCMeta):
    @abstractmethod
    def started(self, value=None) -> None:
        """
        Signal that the task has started.

        :param value: object passed back to the starter of the task
        """


class TaskGroup(metaclass=ABCMeta):
    """
    Groups several asynchronous tasks together.

    :ivar cancel_scope: the cancel scope inherited by all child tasks
    :vartype cancel_scope: CancelScope
    """

    cancel_scope: 'CancelScope'

    @abstractmethod
    def spawn(self, func: Callable[..., Coroutine], *args, name=None) -> DeprecatedAwaitable:
        """
        Launch a new task in this task group.

        :param func: a coroutine function
        :param args: positional arguments to call the function with
        :param name: name of the task, for the purposes of introspection and debugging
        """

    @abstractmethod
    async def start(self, func: Callable[..., Coroutine], *args, name=None) -> None:
        """
        Launch a new task and wait until it signals for readiness.

        :param func: a coroutine function
        :param args: positional arguments to call the function with
        :param name: name of the task, for the purposes of introspection and debugging
        :return: the value passed to ``task_status.started()``
        :raises RuntimeError: if the task finishes without calling ``task_status.started()``

        .. versionadded:: 3.0
        """

    @abstractmethod
    async def __aenter__(self) -> 'TaskGroup':
        """Enter the task group context and allow starting new tasks."""

    @abstractmethod
    async def __aexit__(self, exc_type: Optional[Type[BaseException]],
                        exc_val: Optional[BaseException],
                        exc_tb: Optional[TracebackType]) -> Optional[bool]:
        """Exit the task group context waiting for all tasks to finish."""



import os
import platform
import sys
import time
from functools import partial
from subprocess import CalledProcessError
from textwrap import dedent

import pytest

from anyio import (
    CancelScope, create_task_group, fail_after, open_process, run_process, run_sync_in_process,
    wait_all_tasks_blocked)
from anyio._core._subprocesses import _process_pool_workers
from anyio.streams.buffered import BufferedByteReceiveStream

pytestmark = pytest.mark.anyio


@pytest.fixture(autouse=True)
def check_compatibility(anyio_backend_name):
    if anyio_backend_name == 'asyncio':
        if platform.system() == 'Windows' and sys.version_info < (3, 8):
            pytest.skip('Python < 3.8 uses SelectorEventLoop by default and it does not support '
                        'subprocesses')


@pytest.mark.parametrize('shell, command', [
    pytest.param(True, f'{sys.executable} -c "import sys; print(sys.stdin.read()[::-1])"',
                 id='shell'),
    pytest.param(False, [sys.executable, '-c', 'import sys; print(sys.stdin.read()[::-1])'],
                 id='exec')
])
async def test_run_process(shell, command, anyio_backend_name):
    process = await run_process(command, input=b'abc')
    assert process.returncode == 0
    assert process.stdout.rstrip() == b'cba'


async def test_run_process_checked():
    with pytest.raises(CalledProcessError) as exc:
        await run_process([sys.executable, '-c',
                           'import sys; print("stderr-text", file=sys.stderr); '
                           'print("stdout-text"); sys.exit(1)'], check=True)

    assert exc.value.returncode == 1
    assert exc.value.stdout.rstrip() == b'stdout-text'
    assert exc.value.stderr.rstrip() == b'stderr-text'


@pytest.mark.skipif(platform.system() == 'Windows',
                    reason='process.terminate() kills the process instantly on Windows')
async def test_terminate(tmp_path):
    script_path = tmp_path / 'script.py'
    script_path.write_text(dedent("""\
        import signal, sys, time

        def terminate(signum, frame):
            sys.exit(2)

        signal.signal(signal.SIGTERM, terminate)
        print('ready', flush=True)
        time.sleep(5)
    """))
    async with await open_process([sys.executable, str(script_path)]) as process:
        buffered_stdout = BufferedByteReceiveStream(process.stdout)
        line = await buffered_stdout.receive_until(b'\n', 100)
        assert line.rstrip() == b'ready'

        process.terminate()
        assert await process.wait() == 2


class TestProcessPool:
    async def test_run_sync_in_process_pool(self):
        """
        Test that the function runs in a different process, and the same process in both calls.

        """
        worker_pid = await run_sync_in_process(os.getpid)
        assert worker_pid != os.getpid()
        assert await run_sync_in_process(os.getpid) == worker_pid

    async def test_identical_sys_path(self):
        """Test that partial() can be used to pass keyword arguments."""
        assert await run_sync_in_process(eval, 'sys.path') == sys.path

    async def test_partial(self):
        """Test that partial() can be used to pass keyword arguments."""
        assert await run_sync_in_process(partial(sorted, reverse=True), ['a', 'b']) == ['b', 'a']

    async def test_exception(self):
        """Test that exceptions are delivered properly."""
        with pytest.raises(ValueError, match='invalid literal for int'):
            assert await run_sync_in_process(int, 'a')

    async def test_print(self):
        """Test that print() won't interfere with parent-worker communication."""
        worker_pid = await run_sync_in_process(os.getpid)
        await run_sync_in_process(print, 'hello')
        await run_sync_in_process(print, 'world')
        assert await run_sync_in_process(os.getpid) == worker_pid

    async def test_cancel_before(self):
        """
        Test that starting run_sync_in_process() in a cancelled scope does not cause a worker
        process to be reserved.

        """
        with CancelScope() as scope:
            scope.cancel()
            await run_sync_in_process(os.getpid)

        pytest.raises(LookupError, _process_pool_workers.get)

    async def test_cancel_during(self):
        """
        Test that cancelling an operation on the worker process causes the process to be killed.

        """
        worker_pid = await run_sync_in_process(os.getpid)
        with fail_after(4):
            async with create_task_group() as tg:
                tg.spawn(partial(run_sync_in_process, cancellable=True), time.sleep, 5)
                await wait_all_tasks_blocked()
                tg.cancel_scope.cancel()

        # The previous worker was killed so we should get a new one now
        assert await run_sync_in_process(os.getpid) != worker_pid



__all__ = ('maybe_async', 'maybe_async_cm', 'run', 'sleep', 'current_time', 'get_all_backends',
           'get_cancelled_exc_class', 'BrokenResourceError', 'BrokenWorkerProcess',
           'BusyResourceError', 'ClosedResourceError', 'DelimiterNotFound', 'EndOfStream',
           'ExceptionGroup', 'IncompleteRead', 'TypedAttributeLookupError', 'WouldBlock',
           'AsyncFile', 'open_file', 'aclose_forcefully', 'open_signal_receiver', 'connect_tcp',
           'connect_unix', 'create_tcp_listener', 'create_unix_listener', 'create_udp_socket',
           'create_connected_udp_socket', 'getaddrinfo', 'getnameinfo', 'wait_socket_readable',
           'wait_socket_writable', 'create_memory_object_stream', 'run_process', 'open_process',
           'current_default_worker_process_limiter', 'run_sync_in_process', 'create_lock',
           'CapacityLimiter', 'CapacityLimiterStatistics', 'Condition', 'ConditionStatistics',
           'Event', 'EventStatistics', 'Lock', 'LockStatistics', 'Semaphore',
           'SemaphoreStatistics', 'create_condition', 'create_event', 'create_semaphore',
           'create_capacity_limiter', 'open_cancel_scope', 'fail_after', 'move_on_after',
           'current_effective_deadline', 'TASK_STATUS_IGNORED', 'CancelScope', 'create_task_group',
           'TaskInfo', 'get_current_task', 'get_running_tasks', 'wait_all_tasks_blocked',
           'run_sync_in_worker_thread', 'run_async_from_thread', 'run_sync_from_thread',
           'current_default_worker_thread_limiter', 'create_blocking_portal',
           'start_blocking_portal', 'typed_attribute', 'TypedAttributeSet',
           'TypedAttributeProvider')

from ._core._compat import maybe_async, maybe_async_cm
from ._core._eventloop import current_time, get_all_backends, get_cancelled_exc_class, run, sleep
from ._core._exceptions import (
    BrokenResourceError, BrokenWorkerProcess, BusyResourceError, ClosedResourceError,
    DelimiterNotFound, EndOfStream, ExceptionGroup, IncompleteRead, TypedAttributeLookupError,
    WouldBlock)
from ._core._fileio import AsyncFile, open_file
from ._core._resources import aclose_forcefully
from ._core._signals import open_signal_receiver
from ._core._sockets import (
    connect_tcp, connect_unix, create_connected_udp_socket, create_tcp_listener, create_udp_socket,
    create_unix_listener, getaddrinfo, getnameinfo, wait_socket_readable, wait_socket_writable)
from ._core._streams import create_memory_object_stream
from ._core._subprocesses import (
    current_default_worker_process_limiter, open_process, run_process, run_sync_in_process)
from ._core._synchronization import (
    CapacityLimiter, CapacityLimiterStatistics, Condition, ConditionStatistics, Event,
    EventStatistics, Lock, LockStatistics, Semaphore, SemaphoreStatistics, create_capacity_limiter,
    create_condition, create_event, create_lock, create_semaphore)
from ._core._tasks import (
    TASK_STATUS_IGNORED, CancelScope, create_task_group, current_effective_deadline, fail_after,
    move_on_after, open_cancel_scope)
from ._core._testing import TaskInfo, get_current_task, get_running_tasks, wait_all_tasks_blocked
from ._core._threads import (
    create_blocking_portal, current_default_worker_thread_limiter, run_async_from_thread,
    run_sync_from_thread, run_sync_in_worker_thread, start_blocking_portal)
from ._core._typedattr import TypedAttributeProvider, TypedAttributeSet, typed_attribute

# Re-export imports so they look like they live directly in this package
for key, value in list(locals().items()):
    if getattr(value, '__module__', '').startswith('anyio.'):
        value.__module__ = __name__



import pytest

from anyio import create_task_group, run
from anyio.lowlevel import RunVar, cancel_shielded_checkpoint, checkpoint, checkpoint_if_cancelled

pytestmark = pytest.mark.anyio


@pytest.mark.parametrize('cancel', [False, True])
async def test_checkpoint_if_cancelled(cancel):
    async def func():
        nonlocal finished
        tg.spawn(second_func)
        if cancel:
            tg.cancel_scope.cancel()

        await checkpoint_if_cancelled()
        finished = True

    async def second_func():
        nonlocal second_finished
        assert finished != cancel
        second_finished = True

    finished = second_finished = False
    async with create_task_group() as tg:
        tg.spawn(func)

    assert finished != cancel
    assert second_finished


@pytest.mark.parametrize('cancel', [False, True])
async def test_cancel_shielded_checkpoint(cancel):
    async def func():
        nonlocal finished
        await cancel_shielded_checkpoint()
        finished = True

    async def second_func():
        nonlocal second_finished
        assert not finished
        second_finished = True

    finished = second_finished = False
    async with create_task_group() as tg:
        tg.spawn(func)
        tg.spawn(second_func)
        if cancel:
            tg.cancel_scope.cancel()

    assert finished
    assert second_finished


@pytest.mark.parametrize('cancel', [False, True])
async def test_checkpoint(cancel):
    async def func():
        nonlocal finished
        await checkpoint()
        finished = True

    async def second_func():
        nonlocal second_finished
        assert not finished
        second_finished = True

    finished = second_finished = False
    async with create_task_group() as tg:
        tg.spawn(func)
        tg.spawn(second_func)
        if cancel:
            tg.cancel_scope.cancel()

    assert finished != cancel
    assert second_finished


class TestRunVar:
    def test_get_set(self, anyio_backend_name, anyio_backend_options):
        async def taskfunc(index):
            assert var.get() == index
            var.set(index + 1)

        async def main():
            pytest.raises(LookupError, var.get)
            for i in range(2):
                var.set(i)
                async with create_task_group() as tg:
                    tg.spawn(taskfunc, i)

                assert var.get() == i + 1

        var = RunVar('var')
        for _ in range(2):
            run(main, backend=anyio_backend_name, backend_options=anyio_backend_options)

    async def test_reset_token_used_on_wrong_runvar(self):
        var1 = RunVar('var1')
        var2 = RunVar('var2')
        token = var1.set('blah')
        with pytest.raises(ValueError, match='This token does not belong to this RunVar'):
            var2.reset(token)

    async def test_reset_token_used_twice(self):
        var = RunVar('var')
        token = var.set('blah')
        var.reset(token)
        with pytest.raises(ValueError, match='This token has already been used'):
            var.reset(token)



import os
import pickle
import subprocess
import sys
from collections import deque
from subprocess import DEVNULL, PIPE, CalledProcessError, CompletedProcess
from typing import Callable, Deque, List, Optional, Sequence, Set, Tuple, TypeVar, Union, cast

from ..abc import ByteReceiveStream, ByteSendStream, Process
from ..lowlevel import RunVar, checkpoint_if_cancelled
from ..streams.buffered import BufferedByteReceiveStream
from ._eventloop import current_time, get_asynclib, get_cancelled_exc_class
from ._exceptions import BrokenWorkerProcess
from ._synchronization import CapacityLimiter
from ._tasks import CancelScope, create_task_group, fail_after

T_Retval = TypeVar('T_Retval')
_process_pool_workers: RunVar[Set[Process]] = RunVar('_process_pool_workers')
_process_pool_idle_workers: RunVar[Deque[Tuple[Process, float]]] = RunVar(
    '_process_pool_idle_workers')
_default_process_limiter: RunVar[CapacityLimiter] = RunVar('_default_process_limiter')

WORKER_MAX_IDLE_TIME = 300  # 5 minutes


async def run_process(command: Union[str, Sequence[str]], *, input: Optional[bytes] = None,
                      stdout: int = PIPE, stderr: int = PIPE,
                      check: bool = True) -> CompletedProcess:
    """
    Run an external command in a subprocess and wait until it completes.

    .. seealso:: :func:`subprocess.run`

    :param command: either a string to pass to the shell, or an iterable of strings containing the
        executable name or path and its arguments
    :param input: bytes passed to the standard input of the subprocess
    :param stdout: either :data:`subprocess.PIPE` or :data:`subprocess.DEVNULL`
    :param stderr: one of :data:`subprocess.PIPE`, :data:`subprocess.DEVNULL` or
        :data:`subprocess.STDOUT`
    :param check: if ``True``, raise :exc:`~subprocess.CalledProcessError` if the process
        terminates with a return code other than 0
    :return: an object representing the completed process
    :raises ~subprocess.CalledProcessError: if ``check`` is ``True`` and the process exits with a
        nonzero return code

    """
    async def drain_stream(stream, index):
        chunks = [chunk async for chunk in stream]
        stream_contents[index] = b''.join(chunks)

    async with await open_process(command, stdin=PIPE if input else DEVNULL, stdout=stdout,
                                  stderr=stderr) as process:
        stream_contents = [None, None]
        try:
            async with create_task_group() as tg:
                if process.stdout:
                    tg.spawn(drain_stream, process.stdout, 0)
                if process.stderr:
                    tg.spawn(drain_stream, process.stderr, 1)
                if process.stdin and input:
                    await process.stdin.send(input)
                    await process.stdin.aclose()

                await process.wait()
        except BaseException:
            process.kill()
            raise

    output, errors = stream_contents
    if check and process.returncode != 0:
        raise CalledProcessError(cast(int, process.returncode), command, output, errors)

    return CompletedProcess(command, cast(int, process.returncode), output, errors)


async def open_process(command: Union[str, Sequence[str]], *, stdin: int = PIPE,
                       stdout: int = PIPE, stderr: int = PIPE) -> Process:
    """
    Start an external command in a subprocess.

    .. seealso:: :class:`subprocess.Popen`

    :param command: either a string to pass to the shell, or an iterable of strings containing the
        executable name or path and its arguments
    :param stdin: either :data:`subprocess.PIPE` or :data:`subprocess.DEVNULL`
    :param stdout: either :data:`subprocess.PIPE` or :data:`subprocess.DEVNULL`
    :param stderr: one of :data:`subprocess.PIPE`, :data:`subprocess.DEVNULL` or
        :data:`subprocess.STDOUT`
    :return: an asynchronous process object

    """
    shell = isinstance(command, str)
    return await get_asynclib().open_process(command, shell=shell, stdin=stdin, stdout=stdout,
                                             stderr=stderr)


async def run_sync_in_process(
        func: Callable[..., T_Retval], *args, cancellable: bool = False,
        limiter: Optional[CapacityLimiter] = None) -> T_Retval:
    """
    Call the given function with the given arguments in a worker process.

    If the ``cancellable`` option is enabled and the task waiting for its completion is cancelled,
    the worker process running it will be abruptly terminated using SIGKILL (or
    ``terminateProcess()`` on Windows).

    :param func: a callable
    :param args: positional arguments for the callable
    :param cancellable: ``True`` to allow cancellation of the operation while it's running
    :param limiter: capacity limiter to use to limit the total amount of processes running
        (if omitted, the default limiter is used)
    :return: an awaitable that yields the return value of the function.

    """
    async def send_raw_command(pickled_cmd: bytes):
        try:
            await stdin.send(pickled_cmd)
            response = await buffered.receive_until(b'\n', 50)
            status, length = response.split(b' ')
            if status not in (b'RETURN', b'EXCEPTION'):
                raise RuntimeError(f'Worker process returned unexpected response: {response!r}')

            pickled_response = await buffered.receive_exactly(int(length))
        except BaseException as exc:
            workers.discard(process)
            try:
                process.kill()
                with CancelScope(shield=True):
                    await process.aclose()
            except ProcessLookupError:
                pass

            if isinstance(exc, get_cancelled_exc_class()):
                raise
            else:
                raise BrokenWorkerProcess from exc

        retval = pickle.loads(pickled_response)
        if status == b'EXCEPTION':
            assert isinstance(retval, BaseException)
            raise retval
        else:
            return retval

    # First pickle the request before trying to reserve a worker process
    await checkpoint_if_cancelled()
    request = pickle.dumps(('run', func, args), protocol=pickle.HIGHEST_PROTOCOL)

    # If this is the first run in this event loop thread, set up the necessary variables
    try:
        workers = _process_pool_workers.get()
        idle_workers = _process_pool_idle_workers.get()
    except LookupError:
        workers = set()
        idle_workers = deque()
        _process_pool_workers.set(workers)
        _process_pool_idle_workers.set(idle_workers)
        get_asynclib().setup_process_pool_exit_at_shutdown(workers)

    async with (limiter or current_default_worker_process_limiter()):
        # Pop processes from the pool (starting from the most recently used) until we find one that
        # hasn't exited yet
        process: Process
        while idle_workers:
            process, idle_since = idle_workers.pop()
            if process.returncode is None:
                stdin = cast(ByteSendStream, process.stdin)
                buffered = BufferedByteReceiveStream(cast(ByteReceiveStream, process.stdout))

                # Prune any other workers that have been idle for WORKER_MAX_IDLE_TIME seconds or
                # longer
                now = current_time()
                killed_processes: List[Process] = []
                while idle_workers:
                    if now - idle_workers[0][1] < WORKER_MAX_IDLE_TIME:
                        break

                    process, idle_since = idle_workers.popleft()
                    process.kill()
                    workers.remove(process)
                    killed_processes.append(process)

                with CancelScope(shield=True):
                    for process in killed_processes:
                        await process.aclose()

                break

            workers.remove(process)
        else:
            command = [sys.executable, '-u', '-m', 'anyio._core._subprocess_worker']
            process = await open_process(command, stdin=subprocess.PIPE,
                                         stdout=subprocess.PIPE)
            try:
                stdin = cast(ByteSendStream, process.stdin)
                buffered = BufferedByteReceiveStream(cast(ByteReceiveStream, process.stdout))
                with fail_after(20):
                    message = await buffered.receive(6)

                if message != b'READY\n':
                    raise BrokenWorkerProcess(
                        f'Worker process returned unexpected response: {message!r}')

                main_module_path = getattr(sys.modules['__main__'], '__file__', None)
                pickled = pickle.dumps(('init', sys.path, main_module_path),
                                       protocol=pickle.HIGHEST_PROTOCOL)
                await send_raw_command(pickled)
            except (BrokenWorkerProcess, get_cancelled_exc_class()):
                raise
            except BaseException as exc:
                process.kill()
                raise BrokenWorkerProcess('Error during worker process initialization') from exc

            workers.add(process)

        with CancelScope(shield=not cancellable):
            try:
                return await send_raw_command(request)
            finally:
                if process in workers:
                    idle_workers.append((process, current_time()))


def current_default_worker_process_limiter() -> CapacityLimiter:
    """
    Return the capacity limiter that is used by default to limit the number of worker processes.

    :return: a capacity limiter object

    """
    try:
        return _default_process_limiter.get()
    except LookupError:
        limiter = CapacityLimiter(os.cpu_count() or 2)
        _default_process_limiter.set(limiter)
        return limiter
